/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  instmbr - install an MBR (master boot record) at LBA sector 0    */
/*                                                                   */
/*********************************************************************/

/* Utility bin2hex in ozpd was used to generate the hex values
   from the "compmbr" output */

#include <stdio.h>
#include <stdlib.h>

#include "pos.h"
#include "unused.h"

int main(int argc, char **argv)
{
    static unsigned char mbr[440] = {
    "\xFA\x33\xC0\x8E\xD8\x8E\xC0\xB8\x50\x00\x8E\xD0\xBC\x00\x77\xFB"
    "\xBE\x00\x7C\xBF\x00\x06\xB9\x00\x01\xF3\xA5\xB8\x50\x00\x50\xB8"
    "\x24\x01\x50\xCB\x0E\x1F\x88\x16\xB5\x02\xBE\xBE\x02\xB9\x04\x00"
    "\xF6\x04\x80\x75\x0A\x83\xC6\x10\xE2\xF6\xBD\xBD\x01\xEB\x6F\x80"
    "\x7C\x04\x00\xBD\xD8\x01\x74\x66\x26\xC7\x06\xFE\x7D\x00\x00\xB8"
    "\x00\x00\x8E\xC0\xB4\x08\xBF\x00\x00\xCD\x13\x72\x1F\x80\xE1\x3F"
    "\xB5\x00\x89\x0E\xB1\x02\x8A\xD6\xB6\x00\x83\xC2\x01\x89\x16\xB3"
    "\x02\x8B\x44\x08\x8B\x54\x0A\xE8\xB6\x00\xEB\x09\x8A\x74\x01\x8A"
    "\x4C\x02\x8A\x6C\x03\xB8\x00\x00\x8E\xC0\xBB\x00\x7C\xE8\xD3\x00"
    "\x26\x81\x3E\xFE\x7D\x55\xAA\xBD\x14\x02\x75\x12\x8A\x16\xB5\x02"
    "\x8B\xEE\xB8\x00\x00\x50\xB8\x00\x7C\x50\xCB\xCD\x10\x45\xB4\x0E"
    "\x8A\x46\x00\x3C\x00\x75\xF4\x87\xDB\xFB\xF4\xEB\xFA\x4E\x6F\x20"
    "\x61\x63\x74\x69\x76\x65\x20\x70\x61\x72\x74\x69\x74\x69\x6F\x6E"
    "\x20\x66\x6F\x75\x6E\x64\x21\x00\x41\x63\x74\x69\x76\x65\x20\x70"
    "\x61\x72\x74\x20\x68\x61\x73\x20\x69\x6E\x76\x61\x6C\x69\x64\x20"
    "\x70\x61\x72\x74\x69\x74\x69\x6F\x6E\x20\x74\x79\x70\x65\x21\x00"
    "\x46\x61\x69\x6C\x65\x64\x20\x74\x6F\x20\x72\x65\x61\x64\x20\x56"
    "\x42\x52\x21\x00\x56\x42\x52\x20\x6D\x69\x73\x73\x69\x6E\x67\x20"
    "\x61\x61\x35\x35\x20\x73\x69\x67\x6E\x61\x74\x75\x72\x65\x21\x00"
    "\xF7\x36\xB1\x02\xB5\x00\x8A\xCA\xFE\xC1\x33\xD2\xF7\x36\xB3\x02"
    "\x8A\xF2\x52\x8A\xF0\xB0\x00\x51\xB1\x02\xD3\xE8\x59\x8A\xE6\x5A"
    "\x0B\xC8\xC3\x50\x52\x8A\x16\xB5\x02\xB8\x00\x00\xCD\x13\x72\xF5"
    "\x5A\x58\xC3\x50\x53\x51\x52\x06\x8A\x16\xB5\x02\xB8\x01\x02\xCD"
    "\x13\x73\x05\xE8\xDD\xFF\xEB\xF0\x07\x5A\x59\x5B\x58\xC3\x50\x53"
    "\x8A\xC5\xE8\x0D\x00\x8A\xC1\xE8\x08\x00\xB0\x20\xE8\x18\x00\x5B"
    "\x58\xC3\x8A\xD8\x51\xB1\x04\xD2\xE8\x59\xE8\x02\x00\x8A\xC3\x24"
    "\x0F\x04\x90\x27\x14\x40\x27\x53\xBB\x00\x00\xB4\x0E\xCD\x10\x5B"
    "\xC3\x00\x00\x00\x00\x00\x00\x00"
    };
    unsigned char buf[512];
    int drive;
    unsigned long sect = 0;
    int rc;
    
    if (argc <= 1)
    {
        printf("usage: instmbr <drive>\n");
        printf("example: instmbr 81\n");
        printf("which will install to the second hard disk, typically,\n");
        printf("but not necessarily, the D drive\n");
        printf("80 = first, 81 = second, 82 = third, 83 = fourth\n");
        return (EXIT_FAILURE);
    }
    
    drive = strtol(*(argv + 1), NULL, 16);
    rc = PosAbsoluteDriveRead(drive, sect, 1, buf);

    if (rc != 0)
    {
        printf("read failed with rc %d\n", rc);
        return (EXIT_FAILURE);
    }

    memcpy(buf, mbr, sizeof mbr);

    rc = PosAbsoluteDriveWrite(drive, sect, 1, buf);

    if (rc != 0)
    {
        printf("write failed with rc %d\n", rc);
        return (EXIT_FAILURE);
    }

    printf("MBR has been written to drive %x\n", drive);
    return (0);
}
